@@ -67,6 +67,7 @@ struct gemini {
 	ssize_t			  headers_offs; /* header offset */
 	struct lowdown_buf	**foots; /* footnotes */
 	size_t			  footsz; /* footnotes size  */
+	const char		 *templ; /* output template */
 };
 
 /*
@@ -159,41 +160,6 @@ rndr_link_ref(const struct gemini *st,
 		nl ? " " : "", buf, nl ? "\n" : "");
 }
 
-/*
- * Convert newlines to spaces, elide control characters.
- * If a newline follows a period, it's converted to two spaces.
- * Return zero on failure (memory), non-zero on success.
- */
-static int
-rndr_escape(struct lowdown_buf *out, const char *buf, size_t sz)
-{
-	size_t	 	 i, start = 0;
-	unsigned char	 ch;
-
-	for (i = 0; i < sz; i++) {
-		ch = (unsigned char)buf[i];
-		if (ch == '\n') {
-			if (!hbuf_put(out, buf + start, i - start))
-				return 0;
-			if (out->size && 
-			    out->data[out->size - 1] == '.' &&
-			    !hbuf_putc(out, ' '))
-				return 0;
-			if (!hbuf_putc(out, ' '))
-				return 0;
-			start = i + 1;
-		} else if (ch < 0x80 && iscntrl(ch)) {
-			if (!hbuf_put(out, buf + start, i - start))
-				return 0;
-			start = i + 1;
-		}
-	}
-
-	if (start < sz && !hbuf_put(out, buf + start, sz - start))
-		return 0;
-	return 1;
-}
-
 /*
  * Output optional number of newlines before or after content.
  * Return zero on failure (memory), non-zero on success.
@@ -242,7 +208,7 @@ rndr_buf(struct gemini *st, struct lowdown_buf *out,
 			if (!isspace((unsigned char)in->data[i]))
 				break;
 
-	if (!rndr_escape(out, in->data + i, in->size - i))
+	if (!lowdown_gemini_esc(out, in->data + i, in->size - i, 1))
 		return 0;
 	if (in->size && st->last_blank != 0)
 		st->last_blank = 0;
@@ -288,11 +254,11 @@ rndr_doc_header(struct gemini *st, struct lowdown_buf *out,
 	if (!(st->flags & LOWDOWN_GEMINI_METADATA))
 		return 1;
 	TAILQ_FOREACH(m, mq, entries) {
-		if (!rndr_escape(out, m->key, strlen(m->key)))
+		if (!lowdown_gemini_esc(out, m->key, strlen(m->key), 1))
 			return 0;
 		if (!HBUF_PUTSL(out, ": "))
 			return 0;
-		if (!rndr_escape(out, m->value, strlen(m->value)))
+		if (!lowdown_gemini_esc(out, m->value, strlen(m->value), 1))
 			return 0;
 		st->last_blank = 0;
 		if (!rndr_buf_vspace(st, out, 1))
@@ -310,12 +276,10 @@ static int
 rndr_meta(struct gemini *st, 
 	const struct lowdown_node *n, struct lowdown_metaq *mq)
 {
-	ssize_t				 last_blank;
-	struct lowdown_buf		*tmp = NULL;
-	struct lowdown_meta		*m;
-	const struct lowdown_node	*child;
-	ssize_t				 val;
-	const char			*ep;
+	ssize_t			 last_blank;
+	struct lowdown_meta	*m;
+	ssize_t			 val;
+	const char		*ep;
 
 	/*
 	 * Manually render the children of the meta into a
@@ -327,43 +291,21 @@ rndr_meta(struct gemini *st,
 	last_blank = st->last_blank;
 	st->last_blank = -1;
 
-	if ((tmp = hbuf_new(128)) == NULL)
-		goto err;
-	if ((m = calloc(1, sizeof(struct lowdown_meta))) == NULL)
-		goto err;
-	TAILQ_INSERT_TAIL(mq, m, entries);
-
-	m->key = strndup(n->rndr_meta.key.data,
-		n->rndr_meta.key.size);
-	if (m->key == NULL)
-		goto err;
-
-	TAILQ_FOREACH(child, &n->children, entries)
-		if (!rndr(tmp, mq, st, child))
-			goto err;
-
-	m->value = strndup(tmp->data, tmp->size);
-	if (m->value == NULL)
-		goto err;
+	if ((m = lowdown_get_meta(n, mq)) == NULL)
+		return 0;
 
 	if (strcmp(m->key, "shiftheadinglevelby") == 0) {
-		val = (ssize_t)strtonum
-			(m->value, -100, 100, &ep);
+		val = (ssize_t)strtonum(m->value, -100, 100, &ep);
 		if (ep == NULL)
 			st->headers_offs = val + 1;
 	} else if (strcmp(m->key, "baseheaderlevel") == 0) {
-		val = (ssize_t)strtonum
-			(m->value, 1, 100, &ep);
+		val = (ssize_t)strtonum(m->value, 1, 100, &ep);
 		if (ep == NULL)
 			st->headers_offs = val;
 	}
 
-	hbuf_free(tmp);
 	st->last_blank = last_blank;
 	return 1;
-err:
-	hbuf_free(tmp);
-	return 0;
 }
 
 /*
@@ -1030,20 +972,29 @@ rndr(struct lowdown_buf *ob, struct lowdown_metaq *mq,
 }
 
 int
-lowdown_gemini_rndr(struct lowdown_buf *ob,
-	void *arg, const struct lowdown_node *n)
+lowdown_gemini_rndr(struct lowdown_buf *ob, void *arg,
+    const struct lowdown_node *n)
 {
 	struct gemini		*st = arg;
-	int			 rc;
+	int			 rc = 0;
 	size_t			 i;
 	struct lowdown_metaq	 metaq;
+	struct lowdown_buf	*tmp = NULL;
 
 	TAILQ_INIT(&metaq);
 	st->last_blank = 0;
 	st->headers_offs = 1;
 
-	rc = rndr(ob, &metaq, st, n);
+	if (st->templ != NULL) {
+		if ((tmp = hbuf_new(64)) == NULL)
+			goto out;
+		if (!rndr(tmp, &metaq, st, n))
+			goto out;
+		rc = lowdown_template(st->templ, tmp, ob, &metaq);
+	} else
+		rc = rndr(ob, &metaq, st, n);
 
+out:
 	link_freeq(&st->linkq);
 	st->linkqsz = 0;
 	st->nolinkqsz = 0;
@@ -1051,6 +1002,7 @@ lowdown_gemini_rndr(struct lowdown_buf *ob,
 	for (i = 0; i < st->footsz; i++)
 		hbuf_free(st->foots[i]);
 
+	hbuf_free(tmp);
 	free(st->foots);
 	st->footsz = 0;
 	st->foots = NULL;
@@ -1068,6 +1020,7 @@ lowdown_gemini_new(const struct lowdown_opts *opts)
 
 	TAILQ_INIT(&p->linkq);
 	p->flags = opts != NULL ? opts->oflags : 0;
+	p->templ = opts != NULL ? opts->templ : NULL;
 
 	/* Only use one kind of flag output. */
 
