@@ -1,9 +1,8 @@
-/*	$Id$ */
 /*
  * Copyright (c) 2008, Natacha Porté
  * Copyright (c) 2011, Vicent Martí
  * Copyright (c) 2014, Xavier Mendez, Devin Torres and the Hoedown authors
- * Copyright (c) 2016--2021 Kristaps Dzonsons <kristaps@bsd.lv>
+ * Copyright (c) Kristaps Dzonsons <kristaps@bsd.lv>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -42,9 +41,9 @@ struct 	html {
 	struct hentryq	 	  headers_used; /* headers we've seen */
 	ssize_t			  headers_offs; /* header offset */
 	unsigned int 		  flags; /* "oflags" in lowdown_opts */
-	int			  noescape; /* don't escape text */
 	struct lowdown_buf	**foots; /* footnotes */
 	size_t			  footsz; /* footnotes size  */
+	const char		 *templ; /* output template */
 };
 
 /*
@@ -52,12 +51,11 @@ struct 	html {
  * Return zero on failure, non-zero on success.
  */
 static int
-escape_html(struct lowdown_buf *ob, const char *source,
-	size_t length, const struct html *st)
+escape_html(struct lowdown_buf *ob, const char *source, size_t length,
+    const struct html *st)
 {
 
-	assert(st->noescape == 0);
-	return hesc_html(ob, source, length, 
+	return lowdown_html_esc(ob, source, length,
 		st->flags & LOWDOWN_HTML_OWASP, 0,
 		st->flags & LOWDOWN_HTML_NUM_ENT);
 }
@@ -66,13 +64,11 @@ escape_html(struct lowdown_buf *ob, const char *source,
  * See escape_html().
  */
 static int
-escape_htmlb(struct lowdown_buf *ob, 
-	const struct lowdown_buf *in, const struct html *st)
+escape_htmlb(struct lowdown_buf *ob,
+    const struct lowdown_buf *in, const struct html *st)
 {
 
-	return st->noescape ?
-		hbuf_putb(ob, in) :
-		escape_html(ob, in->data, in->size, st);
+	return escape_html(ob, in->data, in->size, st);
 }
 
 /*
@@ -81,12 +77,11 @@ escape_htmlb(struct lowdown_buf *ob,
  * Return zero on failure, non-zero on success.
  */
 static int
-escape_literal(struct lowdown_buf *ob, 
-	const struct lowdown_buf *in, const struct html *st)
+escape_literal(struct lowdown_buf *ob,
+    const struct lowdown_buf *in, const struct html *st)
 {
 
-	assert(st->noescape == 0);
-	return hesc_html(ob, in->data, in->size, 
+	return lowdown_html_esc(ob, in->data, in->size,
 		st->flags & LOWDOWN_HTML_OWASP, 1,
 		st->flags & LOWDOWN_HTML_NUM_ENT);
 }
@@ -97,11 +92,10 @@ escape_literal(struct lowdown_buf *ob,
  */
 static int
 escape_href(struct lowdown_buf *ob, const struct lowdown_buf *in,
-	const struct html *st)
+    const struct html *st)
 {
 
-	assert(st->noescape == 0);
-	return hesc_href(ob, in->data, in->size);
+	return lowdown_html_esc_href(ob, in->data, in->size);
 }
 
 /*
@@ -112,7 +106,7 @@ static int
 escape_attr(struct lowdown_buf *ob, const struct lowdown_buf *in)
 {
 
-	return hesc_attr(ob, in->data, in->size);
+	return lowdown_html_esc_attr(ob, in->data, in->size);
 }
 
 static int
@@ -125,9 +119,8 @@ newline(struct lowdown_buf *ob)
 }
 
 static int
-rndr_autolink(struct lowdown_buf *ob, 
-	const struct rndr_autolink *parm,
-	const struct html *st)
+rndr_autolink(struct lowdown_buf *ob,
+    const struct rndr_autolink *parm, const struct html *st)
 {
 
 	if (parm->link.size == 0)
@@ -149,8 +142,8 @@ rndr_autolink(struct lowdown_buf *ob,
 	 */
 
 	if (hbuf_strprefix(&parm->link, "mailto:")) {
-		if (!escape_html(ob, 
-		    parm->link.data + 7, 
+		if (!escape_html(ob,
+		    parm->link.data + 7,
 		    parm->link.size - 7, st))
 			return 0;
 	} else {
@@ -162,9 +155,8 @@ rndr_autolink(struct lowdown_buf *ob,
 }
 
 static int
-rndr_blockcode(struct lowdown_buf *ob, 
-	const struct rndr_blockcode *parm,
-	const struct html *st)
+rndr_blockcode(struct lowdown_buf *ob,
+    const struct rndr_blockcode *parm, const struct html *st)
 {
 	if (!newline(ob))
 		return 0;
@@ -188,7 +180,7 @@ rndr_blockcode(struct lowdown_buf *ob,
 
 static int
 rndr_definition_data(struct lowdown_buf *ob,
-	const struct lowdown_buf *content)
+    const struct lowdown_buf *content)
 {
 
 	if (!HBUF_PUTSL(ob, "<dd>\n"))
@@ -200,7 +192,7 @@ rndr_definition_data(struct lowdown_buf *ob,
 
 static int
 rndr_definition_title(struct lowdown_buf *ob,
-	const struct lowdown_buf *content)
+    const struct lowdown_buf *content)
 {
 	size_t	 sz;
 
@@ -217,7 +209,7 @@ rndr_definition_title(struct lowdown_buf *ob,
 
 static int
 rndr_definition(struct lowdown_buf *ob,
-	const struct lowdown_buf *content)
+    const struct lowdown_buf *content)
 {
 
 	if (!newline(ob))
@@ -230,23 +222,70 @@ rndr_definition(struct lowdown_buf *ob,
 }
 
 static int
-rndr_blockquote(struct lowdown_buf *ob,
-	const struct lowdown_buf *content)
+rndr_blockquote(const struct html *st,
+    const struct rndr_blockquote *param, struct lowdown_buf *ob,
+    const struct lowdown_buf *content)
 {
+	size_t	 i;
 
 	if (!newline(ob))
 		return 0;
-	if (!HBUF_PUTSL(ob, "<blockquote>\n"))
+	if (param->type == BLOCKQUOTE_REGULAR || !(st->flags &
+	    (LOWDOWN_HTML_CALLOUT_GFM|LOWDOWN_HTML_CALLOUT_MDN)))
+		return HBUF_PUTSL(ob, "<blockquote>\n") &&
+		    hbuf_putb(ob, content) &&
+		    HBUF_PUTSL(ob, "</blockquote>\n");
+
+	if (!HBUF_PUTSL(ob, "<div class=\""))
 		return 0;
-	if (!hbuf_putb(ob, content))
+
+	/* MDN/GFM style blockquotes. */
+
+	if (st->flags & LOWDOWN_HTML_CALLOUT_MDN) {
+		if (!hbuf_printf(ob, "%s",
+		    (param->admonition == ADMONITION_NOTE) ?
+		     "notecard note" :
+		    (param->admonition == ADMONITION_WARNING) ?
+		     "notecard warning" : "callout"))
+			return 0;
+	}
+	if (st->flags & LOWDOWN_HTML_CALLOUT_GFM) {
+		if (!hbuf_printf(ob, "%smarkdown-alert ",
+		    (st->flags & LOWDOWN_HTML_CALLOUT_MDN) ? " " : ""))
+			return 0;
+		if (!hbuf_printf(ob, "markdown-alert-%s",
+		    (param->admonition == ADMONITION_NOTE) ?
+		    "note" :
+		    (param->admonition == ADMONITION_WARNING) ?
+		    "warning" : "callout"))
+			return 0;
+	}
+
+	if (!HBUF_PUTSL(ob, "\">\n"))
 		return 0;
-	return HBUF_PUTSL(ob, "</blockquote>\n");
+
+	/*
+	 * Weird: content callouts have their initial callout type
+	 * removed.  Fortunately, this is hard-coded, so we know its
+	 * exact length.  Remove the callout type and any extra spacing
+	 * that follows it.
+	 */
+
+	if (param->admonition == ADMONITION_CALLOUT && content->size > 28) {
+		i = 28;
+		while (i < content->size && content->data[i] == ' ')
+			i++;
+		if (!HBUF_PUTSL(ob, "<p>") ||
+		    !hbuf_put(ob, content->data + i, content->size - i))
+			return 0;
+	} else if (!hbuf_putb(ob, content))
+		return 0;
+	return HBUF_PUTSL(ob, "</div>\n");
 }
 
 static int
 rndr_codespan(struct lowdown_buf *ob,
-	const struct rndr_codespan *param, 
-	const struct html *st)
+    const struct rndr_codespan *param, const struct html *st)
 {
 
 	if (!HBUF_PUTSL(ob, "<code>"))
@@ -258,7 +297,7 @@ rndr_codespan(struct lowdown_buf *ob,
 
 static int
 rndr_strikethrough(struct lowdown_buf *ob,
-	const struct lowdown_buf *content)
+    const struct lowdown_buf *content)
 {
 
 	if (!HBUF_PUTSL(ob, "<del>"))
@@ -282,7 +321,7 @@ rndr_double_emphasis(struct lowdown_buf *ob,
 
 static int
 rndr_emphasis(struct lowdown_buf *ob,
-	const struct lowdown_buf *content)
+    const struct lowdown_buf *content)
 {
 
 	if (!HBUF_PUTSL(ob, "<em>"))
@@ -294,7 +333,7 @@ rndr_emphasis(struct lowdown_buf *ob,
 
 static int
 rndr_highlight(struct lowdown_buf *ob,
-	const struct lowdown_buf *content)
+    const struct lowdown_buf *content)
 {
 
 	if (!HBUF_PUTSL(ob, "<mark>"))
@@ -313,7 +352,7 @@ rndr_linebreak(struct lowdown_buf *ob)
 
 static int
 rndr_header(struct lowdown_buf *ob, const struct lowdown_buf *content,
-	const struct lowdown_node *n, struct html *st)
+    const struct lowdown_node *n, struct html *st)
 {
 	ssize_t				 level;
 	const struct lowdown_buf	*buf;
@@ -377,10 +416,8 @@ rndr_header(struct lowdown_buf *ob, const struct lowdown_buf *content,
 }
 
 static int
-rndr_link(struct lowdown_buf *ob,
-	const struct lowdown_buf *content,
-	const struct rndr_link *param,
-	const struct html *st)
+rndr_link(struct lowdown_buf *ob, const struct lowdown_buf *content,
+    const struct rndr_link *param, const struct html *st)
 {
 
 	if (!HBUF_PUTSL(ob, "<a href=\"") ||
@@ -409,16 +446,15 @@ rndr_link(struct lowdown_buf *ob,
 }
 
 static int
-rndr_list(struct lowdown_buf *ob,
-	const struct lowdown_buf *content,
-	const struct rndr_list *param)
+rndr_list(struct lowdown_buf *ob, const struct lowdown_buf *content,
+    const struct rndr_list *param)
 {
 
 	if (!newline(ob))
 		return 0;
 	if (param->flags & HLIST_FL_ORDERED) {
 		if (param->start > 1) {
-			if (!hbuf_printf(ob, 
+			if (!hbuf_printf(ob,
 			    "<ol start=\"%zu\">\n", param->start))
 				return 0;
 		} else {
@@ -438,8 +474,7 @@ rndr_list(struct lowdown_buf *ob,
 
 static int
 rndr_listitem(struct lowdown_buf *ob,
-	const struct lowdown_buf *content,
-	const struct lowdown_node *n)
+    const struct lowdown_buf *content, const struct lowdown_node *n)
 {
 	size_t	 size;
 	int	 blk = 0;
@@ -454,7 +489,7 @@ rndr_listitem(struct lowdown_buf *ob,
 	     n->parent != NULL &&
 	     n->parent->parent != NULL &&
 	     n->parent->parent->type == LOWDOWN_DEFINITION &&
-	     (n->parent->parent->rndr_definition.flags & 
+	     (n->parent->parent->rndr_definition.flags &
 	      HLIST_FL_BLOCK)) ||
 	    (!(n->rndr_listitem.flags & HLIST_FL_DEF) &&
 	     n->parent != NULL &&
@@ -510,15 +545,14 @@ rndr_listitem(struct lowdown_buf *ob,
 
 static int
 rndr_paragraph(struct lowdown_buf *ob,
-	const struct lowdown_buf *content, 
-	struct html *st)
+    const struct lowdown_buf *content, struct html *st)
 {
 	size_t	i = 0, org;
 
 	if (content->size == 0)
 		return 1;
 	while (i < content->size &&
-	       isspace((unsigned char)content->data[i])) 
+	       isspace((unsigned char)content->data[i]))
 		i++;
 	if (i == content->size)
 		return 1;
@@ -531,7 +565,7 @@ rndr_paragraph(struct lowdown_buf *ob,
 	if (st->flags & LOWDOWN_HTML_HARD_WRAP) {
 		for ( ; i < content->size; i++) {
 			org = i;
-			while (i < content->size && 
+			while (i < content->size &&
 			       content->data[i] != '\n')
 				i++;
 
@@ -550,7 +584,7 @@ rndr_paragraph(struct lowdown_buf *ob,
 				return 0;
 		}
 	} else {
-		if (!hbuf_put(ob, 
+		if (!hbuf_put(ob,
 		    content->data + i, content->size - i))
 			return 0;
 	}
@@ -560,8 +594,7 @@ rndr_paragraph(struct lowdown_buf *ob,
 
 static int
 rndr_raw_block(struct lowdown_buf *ob,
-	const struct rndr_blockhtml *param,
-	const struct html *st)
+    const struct rndr_blockhtml *param, const struct html *st)
 {
 	size_t	org, sz;
 
@@ -570,9 +603,9 @@ rndr_raw_block(struct lowdown_buf *ob,
 	if ((st->flags & LOWDOWN_HTML_ESCAPE))
 		return escape_htmlb(ob, &param->text, st);
 
-	/* 
+	/*
 	 * FIXME: Do we *really* need to trim the HTML? How does that
-	 * make a difference? 
+	 * make a difference?
 	 */
 
 	sz = param->text.size;
@@ -596,7 +629,7 @@ rndr_raw_block(struct lowdown_buf *ob,
 
 static int
 rndr_triple_emphasis(struct lowdown_buf *ob,
-	const struct lowdown_buf *content)
+    const struct lowdown_buf *content)
 {
 
 	if (!HBUF_PUTSL(ob, "<strong><em>"))
@@ -616,9 +649,8 @@ rndr_hrule(struct lowdown_buf *ob)
 }
 
 static int
-rndr_image(struct lowdown_buf *ob,
-	const struct rndr_image *param, 
-	const struct html *st)
+rndr_image(struct lowdown_buf *ob, const struct rndr_image *param,
+    const struct html *st)
 {
 	char		 dimbuf[32];
 	unsigned int	 x, y;
@@ -630,7 +662,7 @@ rndr_image(struct lowdown_buf *ob,
 	 * that as a cap to the size.
 	 */
 
-	if (param->dims.size && 
+	if (param->dims.size &&
 	    param->dims.size < sizeof(dimbuf) - 1) {
 		memset(dimbuf, 0, sizeof(dimbuf));
 		memcpy(dimbuf, param->dims.data, param->dims.size);
@@ -690,8 +722,7 @@ rndr_image(struct lowdown_buf *ob,
 
 static int
 rndr_raw_html(struct lowdown_buf *ob,
-	const struct rndr_raw_html *param,
-	const struct html *st)
+    const struct rndr_raw_html *param, const struct html *st)
 {
 
 	if (st->flags & LOWDOWN_HTML_SKIP_HTML)
@@ -704,7 +735,7 @@ rndr_raw_html(struct lowdown_buf *ob,
 
 static int
 rndr_table(struct lowdown_buf *ob,
-	const struct lowdown_buf *content)
+    const struct lowdown_buf *content)
 {
 
 	if (!newline(ob))
@@ -718,7 +749,7 @@ rndr_table(struct lowdown_buf *ob,
 
 static int
 rndr_table_header(struct lowdown_buf *ob,
-	const struct lowdown_buf *content)
+    const struct lowdown_buf *content)
 {
 
 	if (!newline(ob))
@@ -732,7 +763,7 @@ rndr_table_header(struct lowdown_buf *ob,
 
 static int
 rndr_table_body(struct lowdown_buf *ob,
-	const struct lowdown_buf *content)
+    const struct lowdown_buf *content)
 {
 
 	if (content->size == 0)
@@ -747,8 +778,8 @@ rndr_table_body(struct lowdown_buf *ob,
 }
 
 static int
-rndr_tablerow(struct lowdown_buf *ob,
-	const struct lowdown_buf *content)
+rndr_table_row(struct lowdown_buf *ob,
+    const struct lowdown_buf *content)
 {
 
 	if (!HBUF_PUTSL(ob, "<tr>\n"))
@@ -760,8 +791,8 @@ rndr_tablerow(struct lowdown_buf *ob,
 
 static int
 rndr_tablecell(struct lowdown_buf *ob,
-	const struct lowdown_buf *content,
-	const struct rndr_table_cell *param)
+    const struct lowdown_buf *content,
+    const struct rndr_table_cell *param)
 {
 
 	if (param->flags & HTBL_FL_HEADER) {
@@ -801,20 +832,20 @@ rndr_tablecell(struct lowdown_buf *ob,
 
 static int
 rndr_superscript(struct lowdown_buf *ob,
-	const struct lowdown_buf *content)
+    const struct lowdown_buf *content, enum lowdown_rndrt type)
 {
+	const char	*elem;
 
-	if (!HBUF_PUTSL(ob, "<sup>"))
-		return 0;
-	if (!hbuf_putb(ob, content))
-		return 0;
-	return HBUF_PUTSL(ob, "</sup>");
+	elem = (type == LOWDOWN_SUPERSCRIPT) ? "sup" : "sub";
+
+	return hbuf_printf(ob, "<%s>", elem) &&
+	    hbuf_putb(ob, content) &&
+	    hbuf_printf(ob, "</%s>", elem);
 }
 
 static int
 rndr_normal_text(struct lowdown_buf *ob,
-	const struct rndr_normal_text *param,
-	const struct html *st)
+    const struct rndr_normal_text *param, const struct html *st)
 {
 
 	return escape_htmlb(ob, &param->text, st);
@@ -822,7 +853,7 @@ rndr_normal_text(struct lowdown_buf *ob,
 
 static int
 rndr_footnote_def(struct lowdown_buf *ob,
-	const struct lowdown_buf *content, size_t num)
+    const struct lowdown_buf *content, size_t num)
 {
 	size_t	i = 0;
 	int	pfound = 0;
@@ -830,14 +861,14 @@ rndr_footnote_def(struct lowdown_buf *ob,
 	/* Insert anchor at the end of first paragraph block. */
 
 	while ((i + 3) < content->size) {
-		if (content->data[i++] != '<') 
+		if (content->data[i++] != '<')
 			continue;
-		if (content->data[i++] != '/') 
+		if (content->data[i++] != '/')
 			continue;
-		if (content->data[i++] != 'p' && 
-		    content->data[i] != 'P') 
+		if (content->data[i++] != 'p' &&
+		    content->data[i] != 'P')
 			continue;
-		if (content->data[i] != '>') 
+		if (content->data[i] != '>')
 			continue;
 		i -= 3;
 		pfound = 1;
@@ -855,7 +886,7 @@ rndr_footnote_def(struct lowdown_buf *ob,
 		    "&#8617;"
 		    "</a>", num))
 			return 0;
-		if (!hbuf_put(ob, 
+		if (!hbuf_put(ob,
 		    content->data + i, content->size - i))
 			return 0;
 	} else {
@@ -868,7 +899,7 @@ rndr_footnote_def(struct lowdown_buf *ob,
 
 static int
 rndr_footnote_ref(struct lowdown_buf *ob,
-	const struct lowdown_buf *content, struct html *st)
+    const struct lowdown_buf *content, struct html *st)
 {
 	void	*pp;
 	size_t	 num = st->footsz + 1;
@@ -887,16 +918,15 @@ rndr_footnote_ref(struct lowdown_buf *ob,
 	if ((st->foots[st->footsz++] = hbuf_dup(content)) == NULL)
 		return 0;
 
-	return hbuf_printf(ob, 
+	return hbuf_printf(ob,
 		"<sup id=\"fnref%zu\">"
 		"<a href=\"#fn%zu\" rel=\"footnote\">"
 		"%zu</a></sup>", num, num, num);
 }
 
 static int
-rndr_math(struct lowdown_buf *ob,
-	const struct rndr_math *param, 
-	const struct html *st)
+rndr_math(struct lowdown_buf *ob, const struct rndr_math *param,
+    const struct html *st)
 {
 
 	if (param->blockmode && !HBUF_PUTSL(ob, "\\["))
@@ -917,71 +947,36 @@ rndr_doc_footer(struct lowdown_buf *ob, const struct html *st)
 {
 	size_t	 i;
 
+	if (st->footsz == 0)
+		return 1;
+
 	/*
 	 * Start by flushing out our footnotes.  Footnotes are "sparse"
 	 * in that we may not have them all defined (?).
 	 */
 
-	if (st->footsz > 0) {
-		if (!newline(ob))
-			return 0;
-		if (!HBUF_PUTSL(ob,
-		    "<div class=\"footnotes\">\n<hr/>\n<ol>\n"))
-			return 0;
-		for (i = 0; i < st->footsz; i++)
-			if (!rndr_footnote_def(ob, st->foots[i], i + 1))
-				return 0;
-		if (!HBUF_PUTSL(ob, "\n</ol>\n</div>\n"))
-			return 0;
-	}
-
-	return (st->flags & LOWDOWN_STANDALONE) ?
-		HBUF_PUTSL(ob, "</body>\n") : 1;
-}
-
-static int
-rndr_root(struct lowdown_buf *ob,
-	const struct lowdown_buf *content,
-	const struct lowdown_metaq *mq,
-	const struct html *st)
-{
-	const struct lowdown_meta	*m;
-	const char			*lang = NULL;
-
-	TAILQ_FOREACH(m, mq, entries)
-		if (strcasecmp(m->key, "lang") == 0)
-			lang = m->value;
-
-	if (st->flags & LOWDOWN_STANDALONE) {
-		if (!HBUF_PUTSL(ob, "<!DOCTYPE html>\n"))
-			return 0;
-		if (lang == NULL) {
-			if (!HBUF_PUTSL(ob, "<html>\n"))
-				return 0;
-		} else {
-			if (!HBUF_PUTSL(ob, "<html lang=\"") ||
-			    !hesc_attr(ob, lang, strlen(lang)) ||
-			    !HBUF_PUTSL(ob, "\">\n"))
-				return 0;
-		}
-	}
-	if (!hbuf_putb(ob, content))
+	if (!newline(ob))
 		return 0;
-	if (!rndr_doc_footer(ob, st))
+	if (!HBUF_PUTSL(ob, "<div class=\"footnotes\">\n<hr/>\n<ol>\n"))
 		return 0;
-	if (st->flags & LOWDOWN_STANDALONE)
-		return HBUF_PUTSL(ob, "</html>\n");
-	return 1;
+	for (i = 0; i < st->footsz; i++)
+		if (!rndr_footnote_def(ob, st->foots[i], i + 1))
+			return 0;
+	return HBUF_PUTSL(ob, "\n</ol>\n</div>\n");
 }
 
 /*
  * Split "val" into multiple strings delimited by two or more whitespace
- * characters, padding the output with "starttag" and "endtag".
+ * characters, padding the output with "starttag" and "endtag".  If
+ * "href" is set, escape as if this were a link; if "attr" were set,
+ * escape as if this were an HTML attribute; if neither, then escape as
+ * if regular HTML.
  * Return zero on failure, non-zero on success.
  */
 static int
-rndr_meta_multi(struct lowdown_buf *ob, const char *b, int href,
-    const char *starttag, const char *endtag)
+rndr_meta_multi(const struct html *st, struct lowdown_buf *ob,
+    const char *b, int href, int attr, const char *starttag,
+    const char *endtag)
 {
 	const char	*start;
 	size_t		 sz, i, bsz;
@@ -1010,9 +1005,13 @@ rndr_meta_multi(struct lowdown_buf *ob, const char *b, int href,
 
 		if (!hbuf_puts(ob, starttag))
 			return 0;
-		if (!href && !hesc_attr(ob, start, sz))
+		if (attr && !lowdown_html_esc_attr(ob, start, sz))
 			return 0;
-		else if (href && !hesc_href(ob, start, sz))
+		else if (href && !lowdown_html_esc_href(ob, start, sz))
+			return 0;
+		else if (!attr && !href && !lowdown_html_esc(ob, start,
+		    sz, st->flags & LOWDOWN_HTML_OWASP, 0,
+		    st->flags & LOWDOWN_HTML_NUM_ENT))
 			return 0;
 		if (!hbuf_puts(ob, endtag))
 			return 0;
@@ -1023,63 +1022,22 @@ rndr_meta_multi(struct lowdown_buf *ob, const char *b, int href,
 	return 1;
 }
 
-/*
- * Allocate a meta-data value on the queue "mq".
- * Return zero on failure, non-zero on success.
- */
 static int
-rndr_meta(struct lowdown_buf *ob,
-	const struct lowdown_buf *content,
-	struct lowdown_metaq *mq,
-	const struct lowdown_node *n, struct html *st)
-{
-	struct lowdown_meta	*m;
-	ssize_t			 val;
-	const char		*ep;
-
-	m = calloc(1, sizeof(struct lowdown_meta));
-	if (m == NULL)
-		return 0;
-	TAILQ_INSERT_TAIL(mq, m, entries);
-
-	m->key = strndup(n->rndr_meta.key.data,
-		n->rndr_meta.key.size);
-	if (m->key == NULL)
-		return 0;
-	m->value = strndup(content->data, content->size);
-	if (m->value == NULL)
-		return 0;
-
-	if (strcmp(m->key, "shiftheadinglevelby") == 0) {
-		val = (ssize_t)strtonum
-			(m->value, -100, 100, &ep);
-		if (ep == NULL)
-			st->headers_offs = val + 1;
-	} else if (strcmp(m->key, "baseheaderlevel") == 0) {
-		val = (ssize_t)strtonum
-			(m->value, 1, 100, &ep);
-		if (ep == NULL)
-			st->headers_offs = val;
-	}
-
-	return 1;
-}
-
-static int
-rndr_doc_header(struct lowdown_buf *ob,
-	const struct lowdown_buf *content,
-	const struct lowdown_metaq *mq, 
-	const struct html *st)
+rndr_root(struct lowdown_buf *ob, const struct lowdown_buf *content,
+    const struct lowdown_metaq *mq, const struct html *st)
 {
 	const struct lowdown_meta	*m;
 	const char			*author = NULL, *title = NULL,
 					*affil = NULL, *date = NULL,
-					*copy = NULL, *rcsauthor = NULL, 
+					*copy = NULL, *rcsauthor = NULL,
 					*rcsdate = NULL, *css = NULL,
-					*script = NULL;
+					*script = NULL, *header = NULL,
+					*lang = NULL;
 
 	if (!(st->flags & LOWDOWN_STANDALONE))
-		return 1;
+		return hbuf_putb(ob, content);
+	if (st->templ != NULL)
+		return lowdown_template(st->templ, content, ob, mq);
 
 	TAILQ_FOREACH(m, mq, entries)
 		if (strcasecmp(m->key, "author") == 0)
@@ -1100,16 +1058,10 @@ rndr_doc_header(struct lowdown_buf *ob,
 			css = m->value;
 		else if (strcasecmp(m->key, "javascript") == 0)
 			script = m->value;
-
-	if (!hbuf_putb(ob, content))
-		return 0;
-
-	if (!HBUF_PUTSL(ob, 
-	    "<head>\n"
-	    "<meta charset=\"utf-8\" />\n"
-	    "<meta name=\"viewport\""
-	    " content=\"width=device-width,initial-scale=1\" />\n"))
-		return 0;
+		else if (strcasecmp(m->key, "htmlheader") == 0)
+			header = m->value;
+		else if (strcasecmp(m->key, "lang") == 0)
+			lang = m->value;
 
 	/* Overrides. */
 
@@ -1118,25 +1070,38 @@ rndr_doc_header(struct lowdown_buf *ob,
 	if (rcsauthor != NULL)
 		author = rcsauthor;
 
-	if (!rndr_meta_multi(ob, affil, 0,
+	if (!HBUF_PUTSL(ob, "<!DOCTYPE html>\n"))
+		return 0;
+	if (lang != NULL &&
+	    (!HBUF_PUTSL(ob, "<html lang=\"") ||
+	     !lowdown_html_esc_attr(ob, lang, strlen(lang)) ||
+	     !HBUF_PUTSL(ob, "\">\n")))
+		return 0;
+	else if (lang == NULL &&
+	    !HBUF_PUTSL(ob, "<html>\n"))
+		return 0;
+	if (!HBUF_PUTSL(ob,
+	    "<head>\n"
+	    "<meta charset=\"utf-8\" />\n"
+	    "<meta name=\"viewport\" content=\""
+	    "width=device-width,initial-scale=1\" />\n"))
+		return 0;
+	if (!rndr_meta_multi(st, ob, affil, 0, 1,
 	    "<meta name=\"creator\" content=\"", "\" />"))
 		return 0;
-	if (!rndr_meta_multi(ob, author, 0,
+	if (!rndr_meta_multi(st, ob, author, 0, 1,
 	    "<meta name=\"author\" content=\"", "\" />"))
 		return 0;
-	if (!rndr_meta_multi(ob, copy, 0,
+	if (!rndr_meta_multi(st, ob, copy, 0, 1,
 	    "<meta name=\"copyright\" content=\"", "\" />"))
 		return 0;
-
 	if (date != NULL) {
 		if (!HBUF_PUTSL(ob, "<meta name=\"date\" "))
 			return 0;
-
 		/*
 		 * Don't use "scheme" if the date isn't in the
 		 * appropriate format.
 		 */
-
 		if (strlen(date) == 10 &&
 		    isdigit((unsigned char)date[0]) &&
 		    isdigit((unsigned char)date[1]) &&
@@ -1150,19 +1115,17 @@ rndr_doc_header(struct lowdown_buf *ob,
 		    isdigit((unsigned char)date[9]) &&
 		    !HBUF_PUTSL(ob, "scheme=\"YYYY-MM-DD\" "))
 			return 0;
-
 		if (!HBUF_PUTSL(ob, "content=\""))
 			return 0;
-		if (!hesc_attr(ob, date, strlen(date)))
+		if (!lowdown_html_esc_attr(ob, date, strlen(date)))
 			return 0;
 		if (!HBUF_PUTSL(ob, "\" />\n"))
 			return 0;
 	}
-
-	if (!rndr_meta_multi(ob, css, 1,
+	if (!rndr_meta_multi(st, ob, css, 1, 0,
 	    "<link rel=\"stylesheet\" href=\"", "\" />"))
 		return 0;
-	if (!rndr_meta_multi(ob, script, 1,
+	if (!rndr_meta_multi(st, ob, script, 1, 0,
 	     "<script src=\"", "\"></script>"))
 		return 0;
 
@@ -1171,39 +1134,115 @@ rndr_doc_header(struct lowdown_buf *ob,
 	if (!HBUF_PUTSL(ob, "<title>"))
 		return 0;
 	if (title != NULL &&
-	    !hesc_html(ob, title, strlen(title),
+	    !lowdown_html_esc(ob, title, strlen(title),
 		    st->flags & LOWDOWN_HTML_OWASP, 0,
 		    st->flags & LOWDOWN_HTML_NUM_ENT))
 		return 0;
-	if (!HBUF_PUTSL(ob, "</title>\n</head>\n<body>\n"))
+	if (!HBUF_PUTSL(ob, "</title>\n"))
 		return 0;
 
+	/* Optional raw HTML header. */
+
+	if (header != NULL) {
+		if (!hbuf_puts(ob, header))
+			return 0;
+		if (header[strlen(header) - 1] != '\n' &&
+		    !HBUF_PUTSL(ob, "\n"))
+			return 0;
+	}
+	if (!HBUF_PUTSL(ob, "</head>\n<body>\n"))
+		return 0;
+
+	if (!hbuf_putb(ob, content))
+		return 0;
+
+	return HBUF_PUTSL(ob, "</body>\n</html>\n");
+}
+
+/*
+ * Allocate a meta-data value on the queue "mq".
+ * Return zero on failure, non-zero on success.
+ */
+static int
+rndr_meta(struct html *st, const struct lowdown_node *n,
+    struct lowdown_metaq *mq)
+{
+	struct lowdown_meta	*m;
+	ssize_t			 val;
+	const char		*ep;
+
+	if ((m = lowdown_get_meta(n, mq)) == NULL)
+		return 0;
+
+	if (strcmp(m->key, "shiftheadinglevelby") == 0) {
+		val = (ssize_t)strtonum(m->value, -100, 100, &ep);
+		if (ep == NULL)
+			st->headers_offs = val + 1;
+	} else if (strcmp(m->key, "baseheaderlevel") == 0) {
+		val = (ssize_t)strtonum(m->value, 1, 100, &ep);
+		if (ep == NULL)
+			st->headers_offs = val;
+	}
+
+	return 1;
+}
+
+/*
+ * Conditionally output a title block for the document.  This needs at
+ * least one of the title, author, or date.
+ */
+static int
+rndr_doc_header(struct lowdown_buf *ob, const struct lowdown_metaq *mq,
+    const struct html *st)
+{
+	const struct lowdown_meta	*m;
+	const char			*author = NULL, *title = NULL,
+					*date = NULL, *rcsauthor = NULL,
+					*rcsdate = NULL;
+
 	if (!(st->flags & LOWDOWN_HTML_TITLEBLOCK))
 		return 1;
+
+	TAILQ_FOREACH(m, mq, entries)
+		if (strcasecmp(m->key, "author") == 0)
+			author = m->value;
+		else if (strcasecmp(m->key, "date") == 0)
+			date = m->value;
+		else if (strcasecmp(m->key, "rcsauthor") == 0)
+			rcsauthor = rcsauthor2str(m->value);
+		else if (strcasecmp(m->key, "rcsdate") == 0)
+			rcsdate = rcsdate2str(m->value);
+		else if (strcasecmp(m->key, "title") == 0)
+			title = m->value;
+
+	/* Overrides. */
+
+	if (rcsdate != NULL)
+		date = rcsdate;
+	if (rcsauthor != NULL)
+		author = rcsauthor;
+
+	/* Conditionally output a title block. */
+
 	if (author == NULL && title == NULL && date == NULL)
 		return 1;
 
-	/*
-	 * Output a title block.  This will be empty if we don't have
-	 * any title components to output.
-	 */
-
 	if (!HBUF_PUTSL(ob, "<header id=\"title-block-header\">\n"))
 		return 0;
 	if (title != NULL &&
 	    (!HBUF_PUTSL(ob, "<h1 class=\"title\">") ||
-	     !hesc_html(ob, title, strlen(title), 
+	     !lowdown_html_esc(ob, title, strlen(title),
 		     st->flags & LOWDOWN_HTML_OWASP, 0,
 		     st->flags & LOWDOWN_HTML_NUM_ENT) ||
 	     !HBUF_PUTSL(ob, "</h1>\n")))
 			return 0;
 	if (author != NULL &&
-	    !rndr_meta_multi(ob, author, 0,
+	    !rndr_meta_multi(st, ob, author, 0, 0,
 		    "<p class=\"author\">", "</p>"))
 		return 0;
 	if (date != NULL &&
 	    (!HBUF_PUTSL(ob, "<p class=\"date\">") ||
-	     !hesc_html(ob, date, strlen(date), 
+	     !lowdown_html_esc(ob, date, strlen(date),
 		     st->flags & LOWDOWN_HTML_OWASP, 0,
 		     st->flags & LOWDOWN_HTML_NUM_ENT) ||
 	     !HBUF_PUTSL(ob, "</p>\n")))
@@ -1212,29 +1251,17 @@ rndr_doc_header(struct lowdown_buf *ob,
 }
 
 static int
-rndr(struct lowdown_buf *ob,
-	struct lowdown_metaq *mq, void *ref, 
-	const struct lowdown_node *n)
+rndr(struct lowdown_buf *ob, struct lowdown_metaq *mq, struct html *st,
+    const struct lowdown_node *n)
 {
 	const struct lowdown_node	*child;
 	struct lowdown_buf		*tmp;
 	int32_t				 ent;
-	struct html			*st = ref;
 	int				 ret = 1, rc = 1;
 
 	if ((tmp = hbuf_new(64)) == NULL)
 		return 0;
 
-	/*
-	 * If we're processing metadata, don't escape the content as we
-	 * read and parse it.  This prevents double-escaping.  We'll
-	 * properly escape things as we inline them (standalone mode) or
-	 * when we write body text.
-	 */
-
-	if (n->type == LOWDOWN_META)
-		st->noescape = 1;
-
 	TAILQ_FOREACH(child, &n->children, entries)
 		if (!rndr(tmp, mq, st, child))
 			goto out;
@@ -1253,13 +1280,14 @@ rndr(struct lowdown_buf *ob,
 
 	switch (n->type) {
 	case LOWDOWN_ROOT:
-		rc = rndr_root(ob, tmp, mq, st);
+		rc = rndr_doc_footer(tmp, st) &&
+			rndr_root(ob, tmp, mq, st);
 		break;
 	case LOWDOWN_BLOCKCODE:
 		rc = rndr_blockcode(ob, &n->rndr_blockcode, st);
 		break;
 	case LOWDOWN_BLOCKQUOTE:
-		rc = rndr_blockquote(ob, tmp);
+		rc = rndr_blockquote(st, &n->rndr_blockquote, ob, tmp);
 		break;
 	case LOWDOWN_DEFINITION:
 		rc = rndr_definition(ob, tmp);
@@ -1271,12 +1299,11 @@ rndr(struct lowdown_buf *ob,
 		rc = rndr_definition_data(ob, tmp);
 		break;
 	case LOWDOWN_DOC_HEADER:
-		rc = rndr_doc_header(ob, tmp, mq, st);
+		rc = rndr_doc_header(ob, mq, st);
 		break;
 	case LOWDOWN_META:
-		st->noescape = 0;
 		if (n->chng != LOWDOWN_CHNG_DELETE)
-			rc = rndr_meta(ob, tmp, mq, n, st);
+			rc = rndr_meta(st, n, mq);
 		break;
 	case LOWDOWN_HEADER:
 		rc = rndr_header(ob, tmp, n, st);
@@ -1303,7 +1330,7 @@ rndr(struct lowdown_buf *ob,
 		rc = rndr_table_body(ob, tmp);
 		break;
 	case LOWDOWN_TABLE_ROW:
-		rc = rndr_tablerow(ob, tmp);
+		rc = rndr_table_row(ob, tmp);
 		break;
 	case LOWDOWN_TABLE_CELL:
 		rc = rndr_tablecell(ob, tmp, &n->rndr_table_cell);
@@ -1342,7 +1369,10 @@ rndr(struct lowdown_buf *ob,
 		rc = rndr_strikethrough(ob, tmp);
 		break;
 	case LOWDOWN_SUPERSCRIPT:
-		rc = rndr_superscript(ob, tmp);
+		rc = rndr_superscript(ob, tmp, n->type);
+		break;
+	case LOWDOWN_SUBSCRIPT:
+		rc = rndr_superscript(ob, tmp, n->type);
 		break;
 	case LOWDOWN_FOOTNOTE:
 		rc = rndr_footnote_ref(ob, tmp, st);
@@ -1402,8 +1432,8 @@ out:
 }
 
 int
-lowdown_html_rndr(struct lowdown_buf *ob,
-	void *arg, const struct lowdown_node *n)
+lowdown_html_rndr(struct lowdown_buf *ob, void *arg,
+    const struct lowdown_node *n)
 {
 	struct html		*st = arg;
 	struct lowdown_metaq	 metaq;
@@ -1436,6 +1466,7 @@ lowdown_html_new(const struct lowdown_opts *opts)
 		return NULL;
 
 	p->flags = opts == NULL ? 0 : opts->oflags;
+	p->templ = opts == NULL ? NULL : opts->templ;
 	return p;
 }
 
